
// gtrace.cpp

#include "stdafx.h"
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include "gtrace.h"



#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



#define INVALID_FILEHANDLE   (UINT)CFile::hFileNull



class GTraceSettings : public CObject
{
public:
   GTraceSettings();

   void Init();

public:
   // settings initialization status
   int m_nStatus;
   
   GTRACEOPTIONS m_options;
   // handle to output file
   UINT m_hFile;

public:
   virtual ~GTraceSettings();
};


// trace initialization status
#define TIS_FAILED   (-1)   // gtrace intialization failed
#define TIS_UNTRIED  0      // settings never loaded
#define TIS_SUCCESS  1      // successful load of settings


// global trace settings
static GTraceSettings _traceSettings;
// buffer size used for formatting output
static int _nBufSize = 1024;



///////////////////////////////////////////////////////////////////
//

GTrace::GTrace(DWORD dwFilter, LPCTSTR lpszFormat, ...)
{
   m_bRun = FALSE;

   if(_traceSettings.m_nStatus != TIS_FAILED && lpszFormat)
   {           
      // if this is the first time, we must initialize global
      // trace information with user configuration
      if(_traceSettings.m_nStatus == TIS_UNTRIED)
      {
         _traceSettings.Init();
      }

      // if initialized ok...
      if(_traceSettings.m_nStatus == TIS_SUCCESS)
      {
         // if no filter is defined (zero), or this filter one of
         // the filters enabled by the user...
         if(!dwFilter || 
                  (_traceSettings.m_options.dwFilter & dwFilter))
         {
            // this will be used later in the destructor
            m_bRun = TRUE;

            // display trace output minus optional arguments
            if(_traceSettings.m_options.dwFlags & TF_NOARGUMENTS)
            {
               m_strOutput += lpszFormat;
            }
            // format full trace output
            else
            {
               LPTSTR lpszBuf = m_strOutput.GetBuffer(_nBufSize);
               if(lpszBuf)
               {
                  va_list argList;
                  va_start(argList, lpszFormat);

                  ::ZeroMemory(lpszBuf, _nBufSize*sizeof(TCHAR));

                  _vsntprintf(lpszBuf, _nBufSize-1, lpszFormat, 
                                 argList);

                  m_strOutput.ReleaseBuffer();

                  va_end(argList);
               }
            }

            // if we're running function timings, record the 
            // current time and delay trace output until destructor
            if(_traceSettings.m_options.dwFlags & 
                                              TF_FUNCTIONTIMINGS)
            {
               _ftime(&m_startTime);
            }
            // if no timings, output now
            else 
            {
               // automatically add a newline character to the end 
               // of our trace output
               m_strOutput += _T("\n");

               Output(m_strOutput, m_strOutput.GetLength());
            }
         }
      }
   }
}



GTrace::~GTrace()
{
   if(m_bRun)
   {
      if(_traceSettings.m_options.dwFlags & TF_FUNCTIONTIMINGS)
      {
         _timeb stopTime;

         // record the current time
         _ftime(&stopTime);
         // find the difference, in milliseconds, of start and 
         // stop times
         long lTotalTime = DiffTime(m_startTime, stopTime);

         // format timings...
         CString strTemp;
         strTemp.Format(_T("...%d"), lTotalTime);

         // trace output is combination of format and 
         // timing strings
         CString strTiming(m_strOutput);
         strTiming += strTemp;

         // append newline character
         strTiming += _T("\n");

         Output(strTiming, strTiming.GetLength());
      }
   }
}


void GTrace::Output(LPCTSTR lpszText, int nLen)
{
   if(lpszText)
   {
      // if our output is directed to the DEBUG window
      if(_traceSettings.m_options.dwFlags & TF_DEBUGOUTPUT)
      {
   #ifdef _DEBUG
         afxDump << lpszText;
   #else
         // if we're in release mode, don't allow any output to 
         // go the DEBUG window..lets not use AfxMessageBox, as we 
         // may still be in app startup
         ::MessageBox(NULL, _T("GTrace can't output to DEBUG" \
                        "window - not in DEBUG"), _T("GTrace"), 
                         MB_OK);
         // turn off tracing
         _traceSettings.m_nStatus = TIS_FAILED;
         m_bRun = FALSE;
   #endif
      }
      // or our output is directed to a file...
      else
      if(_traceSettings.m_hFile != INVALID_FILEHANDLE)
      {
         _write(_traceSettings.m_hFile, (LPVOID)lpszText, nLen);

         // flush to ensure data is written to output... 
         ::FlushFileBuffers((HANDLE)_traceSettings.m_hFile);
      }
   }
}



long GTrace::DiffTime(_timeb startTime, _timeb stopTime)
{
   long lMiliStart = startTime.millitm;
   long lMiliStop = stopTime.millitm;

   lMiliStop += (stopTime.time - startTime.time) * 1000;
   long lTotalTime = (lMiliStop - lMiliStart);

   // return total time in milliseconds
   return lTotalTime;
}








//////////////////////////////////////////////////////////////////
//  GTraceSettings 

GTraceSettings::GTraceSettings()
{
   m_nStatus = TIS_UNTRIED;

   m_options.dwFlags = 0;
   m_options.dwFilter = 0;
   m_hFile = INVALID_FILEHANDLE;
}


// initialize all trace settings
void GTraceSettings::Init()
{
   m_nStatus = TIS_FAILED;

#ifndef _GTRACE_DISABLE_  // just in case

   if(GRegLoadTraceOptions(&m_options))
   {
      if(m_options.dwFlags & TF_ENABLE)
      {
         if(m_options.dwFlags & TF_DEBUGOUTPUT)
         {
            m_nStatus = TIS_SUCCESS;
         }
         // try to open the output file...
         else
         {
            if(!m_options.strFileName.IsEmpty())
            {
               m_hFile = _topen(m_options.strFileName, 
                        O_CREAT|O_TEXT|_O_WRONLY|
                        O_TRUNC, _S_IREAD|_S_IWRITE); 

               if(m_hFile != INVALID_FILEHANDLE)
               {
                  m_nStatus = TIS_SUCCESS;
               }
            }
         }
      }
   }

#endif
}



GTraceSettings::~GTraceSettings()
{
#ifndef _GTRACE_DISABLE_

   if(_traceSettings.m_nStatus == TIS_SUCCESS)
   {
      // if output file is open, close it...
      if(_traceSettings.m_hFile != INVALID_FILEHANDLE)
      {
         _close(_traceSettings.m_hFile);
         _traceSettings.m_hFile = INVALID_FILEHANDLE;
      }

      // set status back to failed
      _traceSettings.m_nStatus = TIS_FAILED;
   }

#endif
}





//////////////////////////////////////////////////////////////////
// get a string entry from the registry

CString GRegQueryString(HKEY hSecKey, LPCTSTR lpszEntry, 
                                            LPCTSTR lpszDefault)
{
   CString strValue;
   DWORD dwType = 0;
   DWORD dwCount = 0;

   if(lpszEntry)
   {
      LONG lResult = ::RegQueryValueEx(hSecKey, (LPTSTR)lpszEntry, 
                                       NULL, &dwType,
                                       NULL, &dwCount);

      if(lResult == ERROR_SUCCESS)
      {
         int nBufLen = dwCount/sizeof(TCHAR);
         LPBYTE lpBuf = (LPBYTE)strValue.GetBuffer(nBufLen);

         lResult = ::RegQueryValueEx(hSecKey, (LPTSTR)lpszEntry, 
                                                NULL, &dwType,
                                                lpBuf, &dwCount);

         strValue.ReleaseBuffer();
      }

      if(lResult != ERROR_SUCCESS && lpszDefault)
      {
         strValue = lpszDefault;
      }
   }
   
   return strValue;
}



//////////////////////////////////////////////////////////////////
// get a dword value from the registry

DWORD GRegQueryDword(HKEY hSecKey, LPCTSTR lpszEntry, 
                                                  DWORD dwDefault)
{
   DWORD dwValue = dwDefault;
   DWORD dwType = 0;
   DWORD dwCount = sizeof(DWORD);

   if(lpszEntry)
   {
      LONG lResult = ::RegQueryValueEx(hSecKey, (LPTSTR)lpszEntry, 
                                       NULL, &dwType, 
                                       (LPBYTE)&dwValue, &dwCount);
   }

   return dwValue;
}




BOOL GRegLoadTraceOptions(LPGTRACEOPTIONS lpTraceOptions)
{
   BOOL bLoad = FALSE;
   HKEY hTrace;

   if(::RegOpenKey(HKEY_CURRENT_USER, GTRACEREG_KEY, &hTrace) 
                                                == ERROR_SUCCESS)
   {
      lpTraceOptions->dwFlags = GRegQueryDword(hTrace, 
                                           GTRACEREG_ENTRY_FLAGS);
      lpTraceOptions->dwFilter = GRegQueryDword(hTrace, 
                                           GTRACEREG_ENTRY_FILTER);
      lpTraceOptions->strFileName = GRegQueryString(hTrace, 
                                       GTRACEREG_ENTRY_FILENAME);

      ::RegCloseKey(hTrace);
      bLoad = TRUE;
   }

   return bLoad;
}



BOOL GRegStoreTraceOptions(LPGTRACEOPTIONS lpTraceOptions)
{
   HKEY hTrace;

   if(::RegOpenKey(HKEY_CURRENT_USER, GTRACEREG_KEY, &hTrace) 
                                                 != ERROR_SUCCESS)
   {
      DWORD dw;
      if(::RegCreateKeyEx(HKEY_CURRENT_USER, GTRACEREG_KEY, 0, 
                              REG_NONE, REG_OPTION_NON_VOLATILE, 
                              KEY_WRITE|KEY_READ, NULL,
                              &hTrace, &dw) != ERROR_SUCCESS)
      {
         return FALSE;
      }
   }

   ::RegSetValueEx(hTrace, GTRACEREG_ENTRY_FLAGS, 0, REG_DWORD, 
                         (const BYTE *)&lpTraceOptions->dwFlags, 
                         (DWORD)sizeof(lpTraceOptions->dwFlags));
   ::RegSetValueEx(hTrace, GTRACEREG_ENTRY_FILTER, 0, REG_DWORD, 
                         (const BYTE *)&lpTraceOptions->dwFilter, 
                         (DWORD)sizeof(lpTraceOptions->dwFilter));
   ::RegSetValueEx(hTrace, GTRACEREG_ENTRY_FILENAME, 0, REG_SZ, 
               (const BYTE *)(LPCTSTR)lpTraceOptions->strFileName, 
               lpTraceOptions->strFileName.GetLength()+1);

   RegCloseKey(hTrace);

   return TRUE;
}

